@@ -1,22 +1,31 @@
 from __future__ import annotations
 
+import dataclasses
 import re
 import sys
-from typing import TYPE_CHECKING, Tuple, Union, cast, overload
+from typing import TYPE_CHECKING, cast
 
 if sys.version_info >= (3, 9):
-    from collections.abc import Iterator, Mapping, MutableMapping, Sequence
+    from collections.abc import Collection, Iterator, Sequence
 else:
-    from typing import Iterator, Mapping, MutableMapping, Sequence
+    from typing import Collection, Iterator, Sequence
 
 if sys.version_info >= (3, 10):
     from types import EllipsisType
 else:
     EllipsisType = type(...)
 
+if sys.version_info >= (3, 12):
+    from typing import override
+else:
+    from typing_extensions import override
+
 import numpy as np
+from multicollections import MultiDict
+from multicollections.abc import MutableMultiMapping, with_default
 from pyparsing import (
     CaselessKeyword,
+    CharsNotIn,
     Combine,
     Dict,
     Forward,
@@ -27,6 +36,7 @@ from pyparsing import (
     Located,
     NoMatch,
     Opt,
+    ParseException,
     ParserElement,
     ParseResults,
     Regex,
@@ -38,11 +48,15 @@ from pyparsing import (
     printables,
 )
 
-from ._types import Data, Dimensioned, DimensionSet, File, StandaloneData, SubDict
-
 if TYPE_CHECKING:
     from numpy.typing import DTypeLike
 
+from ._util import add_to_mapping, as_dict_check_unique
+from .types import Dimensioned, DimensionSet
+
+if TYPE_CHECKING:
+    from ._typing import Data, File, StandaloneData, SubDict
+
 
 def _ascii_numeric_list(
     dtype: DTypeLike,
@@ -281,11 +295,12 @@ def _dict_of(
         keyword_entry = Located(keyword_entry)
 
     dict_ <<= (
-        Literal("{").suppress()
-        + Dict(Group(keyword_entry)[...], asdict=not located)
-        + Literal("}").suppress()
+        Literal("{").suppress() + Group(keyword_entry)[...] + Literal("}").suppress()
     )
 
+    if not located:
+        dict_.set_parse_action(lambda tks: as_dict_check_unique(tks.as_list()))
+
     return dict_
 
 
@@ -347,12 +362,12 @@ _TENSOR = common.ieee_float | (
     + Group(common.ieee_float[3] | common.ieee_float[6] | common.ieee_float[9])
     + Literal(")").suppress()
 ).add_parse_action(lambda tks: np.array(tks[0], dtype=float))
-_PARENTHESIZED = Forward()
-_IDENTIFIER = Combine(Word(_IDENTCHARS, _IDENTBODYCHARS) + Opt(_PARENTHESIZED))
-_PARENTHESIZED <<= Combine(
-    Literal("(")
-    + (_PARENTHESIZED | Word(_IDENTBODYCHARS) + Opt(_PARENTHESIZED))
-    + Literal(")")
+_BALANCED = Forward()
+_BALANCED <<= Opt(CharsNotIn("()")) + Opt(
+    Literal("(") + _BALANCED + Literal(")") + _BALANCED
+)
+_IDENTIFIER = Combine(
+    Word(_IDENTCHARS, _IDENTBODYCHARS) + Opt(Literal("(") + _BALANCED + Literal(")"))
 )
 
 _DIMENSIONED = (Opt(_IDENTIFIER) + _DIMENSIONS + _TENSOR).set_parse_action(
@@ -411,16 +426,17 @@ _FIELD = (Keyword("uniform", _IDENTBODYCHARS).suppress() + _TENSOR) | (
 )
 
 _DIRECTIVE = Word("#", _IDENTBODYCHARS)
-_TOKEN = dbl_quoted_string | _DIRECTIVE | _IDENTIFIER
+_KEYWORD = dbl_quoted_string | _IDENTIFIER
+_TOKEN = _KEYWORD | _DIRECTIVE
 _DATA = Forward()
 _DATA_ENTRY = Forward()
 _KEYWORD_ENTRY = _keyword_entry_of(
-    _TOKEN | _list_of(_IDENTIFIER),
+    _KEYWORD | _list_of(_IDENTIFIER),
     Opt(_DATA, default=""),
     directive=_DIRECTIVE,
     data_entry=_DATA_ENTRY,
 )
-_DICT = _dict_of(_TOKEN, Opt(_DATA, default=""))
+_DICT = _dict_of(_KEYWORD, Opt(_DATA, default=""))
 _LIST_ENTRY = _DICT | _KEYWORD_ENTRY | _DATA_ENTRY
 _LIST = _list_of(_LIST_ENTRY)
 _NUMBER = (
@@ -455,46 +471,9 @@ _STANDALONE_DATA = (
 ).add_parse_action(lambda tks: [None, tks[0]])
 
 
-_FILE = (
-    Dict(_KEYWORD_ENTRY[...] + Opt(Group(_STANDALONE_DATA)) + _KEYWORD_ENTRY[...])
-    .ignore(_COMMENT)
-    .parse_with_tabs()
-)
-
-_DATA_OR_DICT = (_DATA | _DICT).ignore(_COMMENT).parse_with_tabs()
-
-
-@overload
-def loads(s: bytes | str, *, keywords: tuple[()]) -> File | StandaloneData: ...
-
-
-@overload
-def loads(
-    s: bytes | str, *, keywords: tuple[str, ...] | None = None
-) -> File | StandaloneData | Data | SubDict: ...
-
-
-def loads(
-    s: bytes | str, *, keywords: tuple[str, ...] | None = None
-) -> File | StandaloneData | Data | SubDict:
-    if isinstance(s, bytes):
-        s = s.decode("latin-1")
-
-    if keywords == ():
-        data = _FILE.parse_string(s, parse_all=True).as_dict()
-
-        if len(data) == 1 and None in data:
-            data = data[None]
-
-    else:
-        data = _DATA_OR_DICT.parse_string(s, parse_all=True)[0]
-
-    return data
-
-
 _LOCATED_KEYWORD_ENTRIES = Group(
     _keyword_entry_of(
-        _TOKEN,
+        _KEYWORD,
         Opt(_DATA, default=""),
         directive=_DIRECTIVE,
         data_entry=_DATA_ENTRY,
@@ -503,7 +482,7 @@ _LOCATED_KEYWORD_ENTRIES = Group(
 )[...]
 _LOCATED_STANDALONE_DATA = Group(Located(_STANDALONE_DATA))
 
-_LOCATED_FILE = (
+_FILE = (
     Dict(
         _LOCATED_KEYWORD_ENTRIES
         + Opt(_LOCATED_STANDALONE_DATA)
@@ -514,59 +493,83 @@ _LOCATED_FILE = (
 )
 
 
-class Parsed(Mapping[Tuple[str, ...], Union[Data, StandaloneData, EllipsisType]]):
-    def __init__(self, contents: bytes) -> None:
-        self._parsed: MutableMapping[
-            tuple[str, ...],
-            tuple[int, Data | StandaloneData | EllipsisType, int],
-        ] = {}
-        for parse_result in _LOCATED_FILE.parse_string(
-            contents.decode("latin-1"), parse_all=True
-        ):
-            self._parsed.update(self._flatten_result(parse_result))
+class Parsed(
+    MutableMultiMapping["tuple[str, ...]", "Data | StandaloneData | EllipsisType"]
+):
+    @dataclasses.dataclass
+    class _Entry:
+        data: Data | StandaloneData | EllipsisType
+        start: int
+        end: int
+
+    def __init__(self, contents: bytes | str) -> None:
+        if isinstance(contents, bytes):
+            contents_str = contents.decode("latin-1")
+        else:
+            contents_str = contents
+            contents = contents.encode("latin-1")
+
+        try:
+            parse_results = _FILE.parse_string(contents_str, parse_all=True)
+        except ParseException as e:
+            msg = f"Failed to parse contents: {e}"
+            raise ValueError(msg) from e
+        self._parsed = self._flatten_results(parse_results)
 
         self.contents = contents
         self.modified = False
 
     @staticmethod
-    def _flatten_result(
-        parse_result: ParseResults, *, _keywords: tuple[str, ...] = ()
-    ) -> Mapping[
-        tuple[str, ...], tuple[int, Data | StandaloneData | EllipsisType, int]
-    ]:
-        ret: MutableMapping[
-            tuple[str, ...],
-            tuple[int, Data | StandaloneData | EllipsisType, int],
-        ] = {}
-        start = parse_result.locn_start
-        assert isinstance(start, int)
-        item = parse_result.value
-        assert isinstance(item, Sequence)
-        end = parse_result.locn_end
-        assert isinstance(end, int)
-        keyword, *data = item
-        if keyword is None:
-            assert not _keywords
-            assert len(data) == 1
-            assert not isinstance(data[0], ParseResults)
-            ret[()] = (start, data[0], end)
-        else:
-            assert isinstance(keyword, str)
-            ret[(*_keywords, keyword)] = (start, ..., end)
-            for d in data:
-                if isinstance(d, ParseResults):
-                    ret.update(
-                        Parsed._flatten_result(d, _keywords=(*_keywords, keyword))
+    def _flatten_results(
+        parse_results: ParseResults | Sequence[ParseResults],
+        *,
+        _keywords: tuple[str, ...] = (),
+    ) -> MultiDict[tuple[str, ...], Parsed._Entry]:
+        ret: MultiDict[tuple[str, ...], Parsed._Entry] = MultiDict()
+        for parse_result in parse_results:
+            value = parse_result.value
+            assert isinstance(value, Sequence)
+            start = parse_result.locn_start
+            assert isinstance(start, int)
+            end = parse_result.locn_end
+            assert isinstance(end, int)
+            keyword, *data = value
+            if keyword is None:
+                assert not _keywords
+                assert len(data) == 1
+                assert not isinstance(data[0], ParseResults)
+                assert () not in ret
+                ret[()] = Parsed._Entry(data[0], start, end)
+            else:
+                assert isinstance(keyword, str)
+                if len(data) == 0 or isinstance(data[0], ParseResults):
+                    if (*_keywords, keyword) in ret:
+                        msg = f"Duplicate entry found for keyword: {keyword}"
+                        raise ValueError(msg)
+                    ret[(*_keywords, keyword)] = Parsed._Entry(..., start, end)
+                    ret.extend(
+                        Parsed._flatten_results(data, _keywords=(*_keywords, keyword))
                     )
                 else:
-                    ret[(*_keywords, keyword)] = (start, d, end)
+                    if (*_keywords, keyword) in ret and not keyword.startswith("#"):
+                        msg = f"Duplicate entry found for keyword: {keyword}"
+                        raise ValueError(msg)
+                    ret.add((*_keywords, keyword), Parsed._Entry(data[0], start, end))
         return ret
 
-    def __getitem__(
+    @override
+    @with_default
+    def getall(
         self, keywords: tuple[str, ...]
-    ) -> Data | StandaloneData | EllipsisType:
-        _, data, _ = self._parsed[keywords]
-        return data
+    ) -> Collection[Data | StandaloneData | EllipsisType]:
+        return [entry.data for entry in self._parsed.getall(keywords)]
+
+    @override
+    def __setitem__(
+        self, key: tuple[str, ...], value: Data | StandaloneData | EllipsisType
+    ) -> None:  # pragma: no cover
+        msg = "Use 'put' method instead"
+        raise NotImplementedError(msg)
 
     def put(
         self,
@@ -574,87 +577,123 @@ class Parsed(Mapping[Tuple[str, ...], Union[Data, StandaloneData, EllipsisType]]
         data: Data | StandaloneData | EllipsisType,
         content: bytes,
     ) -> None:
-        start, end = self.entry_location(keywords, missing_ok=True)
+        start, end = self.entry_location(keywords)
 
-        diff = len(content) - (end - start)
-        for k, (s, d, e) in self._parsed.items():
-            if s >= end:
-                self._parsed[k] = (s + diff, d, e + diff)
-            elif e > start:
-                self._parsed[k] = (s, d, end + diff)
-            elif s > start:
-                self._parsed[k] = (s, d, e + diff)
+        self._update_content(start, end, content)
+        self._parsed[keywords] = Parsed._Entry(data, start, start + len(content))
+        self._remove_child_entries(keywords)
 
-        self._parsed[keywords] = (start, data, end + diff)
+    @override
+    def add(  # type: ignore[override]
+        self,
+        keywords: tuple[str, ...],
+        data: Data | StandaloneData | EllipsisType,
+        content: bytes,
+    ) -> None:
+        assert keywords not in self._parsed or (
+            keywords and keywords[-1].startswith("#")
+        )
+        assert not keywords or not keywords[-1].startswith("#") or data is not ...
+        assert keywords or data is not ...
 
-        self.contents = self.contents[:start] + content + self.contents[end:]
+        start, end = self.entry_location(keywords, add=True)
+
+        self._parsed.add(keywords, Parsed._Entry(data, start, end))
+        self._update_content(start, end, content)
+
+    @override
+    @with_default
+    def popone(self, keywords: tuple[str, ...]) -> Data | StandaloneData | EllipsisType:
+        start, end = self.entry_location(keywords)
+        entry = self._parsed.popone(keywords)
+        self._remove_child_entries(keywords)
+        self._update_content(start, end, b"")
+        return entry.data
+
+    @override
+    def __contains__(self, keywords: object) -> bool:
+        return keywords in self._parsed
+
+    @override
+    def __iter__(self) -> Iterator[tuple[str, ...]]:
+        return iter(self._parsed)
+
+    @override
+    def __len__(self) -> int:
+        return len(self._parsed)
+
+    def _update_content(self, start: int, end: int, new_content: bytes) -> None:
+        """Update content and adjust positions of other entries."""
+        diff = len(new_content) - (end - start)
+
+        # Update positions of other entries if content length changed
+        if diff != 0:
+            for entry in self._parsed.values():
+                assert isinstance(entry, Parsed._Entry)
+                if entry.start >= end:
+                    entry.start += diff
+                    entry.end += diff
+                elif entry.end > start:
+                    entry.end += diff
+
+        self.contents = self.contents[:start] + new_content + self.contents[end:]
         self.modified = True
 
+    def _remove_child_entries(self, keywords: tuple[str, ...]) -> None:
+        """Remove all child entries of the given keywords."""
         for k in list(self._parsed):
             if keywords != k and keywords == k[: len(keywords)]:
                 del self._parsed[k]
 
-    def __delitem__(self, keywords: tuple[str, ...]) -> None:
-        start, end = self.entry_location(keywords)
-        del self._parsed[keywords]
-
-        for k in list(self._parsed):
-            if keywords == k[: len(keywords)]:
-                del self._parsed[k]
-
-        diff = end - start
-        for k, (s, d, e) in self._parsed.items():
-            if s > end:
-                self._parsed[k] = (s - diff, d, e - diff)
-            elif e > start:
-                self._parsed[k] = (s, d, e - diff)
-
-        self.contents = self.contents[:start] + self.contents[end:]
-        self.modified = True
-
-    def __contains__(self, keywords: object) -> bool:
-        return keywords in self._parsed
-
-    def __iter__(self) -> Iterator[tuple[str, ...]]:
-        return iter(self._parsed)
-
-    def __len__(self) -> int:
-        return len(self._parsed)
-
     def entry_location(
-        self, keywords: tuple[str, ...], *, missing_ok: bool = False
+        self, keywords: tuple[str, ...], *, add: bool = False
     ) -> tuple[int, int]:
-        try:
-            start, _, end = self._parsed[keywords]
-        except KeyError:
-            if missing_ok:
-                if len(keywords) > 1:
-                    assert self[keywords[:-1]] is ...
-                    start, end = self.entry_location(keywords[:-1])
-                    end = self.contents.rindex(b"}", start, end)
-                else:
-                    end = len(self.contents)
-
-                start = end
+        if add or keywords not in self._parsed:
+            if len(keywords) > 1:
+                assert self[keywords[:-1]] is ...
+                start, end = self.entry_location(keywords[:-1])
+                end = self.contents.rindex(b"}", start, end)
             else:
-                raise
+                end = len(self.contents)
+
+            start = end
+        else:
+            entry = self._parsed[keywords]
+            start = entry.start
+            end = entry.end
 
         return start, end
 
     def as_dict(self) -> File:
         ret: File = {}
-        for keywords, (_, data, _) in self._parsed.items():
-            r = ret
-            for k in keywords[:-1]:
-                v = r[k]
-                assert isinstance(v, dict)
-                r = cast("File", v)
-
-            assert isinstance(r, dict)
-            if keywords:
-                r[keywords[-1]] = {} if data is ... else data
+        for keywords, entry in self._parsed.items():
+            assert isinstance(entry, Parsed._Entry)
+            if not keywords:
+                assert entry.data is not ...
+                assert None not in ret
+                ret[None] = entry.data
+            elif entry.data is ...:
+                parent: File | SubDict = ret
+                for k in keywords[:-1]:
+                    sub = parent[k]
+                    assert isinstance(sub, (dict, MultiDict))
+                    parent = sub
+                assert keywords[-1] not in parent
+                parent[keywords[-1]] = {}
             else:
-                assert data is not ...
-                r[None] = data
+                assert entry.data is not ...
+                if len(keywords) == 1:
+                    ret = add_to_mapping(ret, keywords[0], entry.data)
+                else:
+                    grandparent: File | SubDict = ret
+                    for k in keywords[:-2]:
+                        sub = grandparent[k]
+                        assert isinstance(sub, (dict, MultiDict))
+                        grandparent = sub
+                    sub = grandparent[keywords[-2]]
+                    assert isinstance(sub, (dict, MultiDict))
+                    grandparent[keywords[-2]] = add_to_mapping(
+                        sub, keywords[-1], cast("Data", entry.data)
+                    )
 
         return ret
