@@ -1,13 +1,16 @@
-# ruff: noqa: UP006, D100
+# ruff: noqa: D100
 from __future__ import annotations
 
-from pathlib import Path
-from typing import List
+from pathlib import Path  # noqa: TC003
 
-from pydantic import BaseModel
+try:
+    from pydantic import BaseModel
+except ImportError as e:
+    msg = "The preprocessing module requires extra dependencies. Install 'foamlib[preprocessing]' to use it."
+    raise ImportError(msg) from e
 
 from foamlib import FoamCase
-from foamlib.preprocessing.of_dict import FoamDictAssignment
+from foamlib.preprocessing.of_dict import FoamDictAssignment  # noqa: TC001
 
 
 class CaseParameter(BaseModel):
@@ -22,8 +25,8 @@ class CaseModifier(BaseModel):
 
     template_case: Path
     output_case: Path
-    key_value_pairs: List[FoamDictAssignment]
-    case_parameters: List[CaseParameter]
+    key_value_pairs: list[FoamDictAssignment]
+    case_parameters: list[CaseParameter]
 
     def create_case(self) -> FoamCase:
         """Create a new case by copying the template case to the output case directory."""
@@ -39,7 +42,7 @@ class CaseModifier(BaseModel):
         for pair in self.key_value_pairs:
             pair.set_value(case_path=self.output_case)
 
-        with open(self.output_case / "case.json", "w") as json_file:
+        with (self.output_case / "case.json").open("w") as json_file:
             json_file.write(self.model_dump_json(indent=4))
 
         return of_case
